@@ -20,7 +20,7 @@
  */
 
 /* IMPORTANT NOTE: no function in this file do its own connection locking
-   except for pg_execute and pq_fetch (that are somehow high-level. This means
+   except for pg_execute and pq_fetch (that are somehow high-level). This means
    that all the othe functions should be called while holding a lock to the
    connection.
 */
@@ -153,10 +153,15 @@ pq_raise(connectionObject *conn, cursorObject *curs, PGresult *pgres)
     const char *err2 = NULL;
     const char *code = NULL;
 
-    if ((conn == NULL && curs == NULL) || (curs != NULL && conn == NULL)) {
+    if (conn == NULL) {
         PyErr_SetString(Error, "psycopg went psycotic and raised a null error");
         return;
     }
+    
+    /* if the connection has somehow beed broken, we mark the connection
+       object as closed but requiring cleanup */
+    if (conn->pgconn != NULL && PQstatus(conn->pgconn) == CONNECTION_BAD)
+        conn->closed = 2;
 
     if (pgres == NULL && curs != NULL)
         pgres = curs->pgres;
@@ -326,7 +331,7 @@ pq_execute_command_locked(connectionObject *conn, const char *query,
 
     retvalue = 0;
     IFCLEARPGRES(*pgres);
-
+    
  cleanup:
     return retvalue;
 }
@@ -348,7 +353,6 @@ pq_complete_error(connectionObject *conn, PGresult **pgres, char **error)
         pq_raise(conn, NULL, *pgres);
     else if (*error != NULL) {
         PyErr_SetString(OperationalError, *error);
-        free(*error);
     } else {
         PyErr_SetString(OperationalError, "unknown error");
     }
@@ -368,7 +372,7 @@ pq_complete_error(connectionObject *conn, PGresult **pgres, char **error)
    On error, -1 is returned, and the pgres argument will hold the
    relevant result structure.
  */
-static int
+int
 pq_begin_locked(connectionObject *conn, PGresult **pgres, char **error)
 {
     const char *query[] = {
@@ -416,12 +420,15 @@ pq_commit(connectionObject *conn)
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&conn->lock);
+    conn->mark += 1;
 
     pq_clear_async(conn);
     retvalue = pq_execute_command_locked(conn, "COMMIT", &pgres, &error);
 
     pthread_mutex_unlock(&conn->lock);
     Py_END_ALLOW_THREADS;
+    
+    conn_notice_process(conn);
 
     if (retvalue < 0)
         pq_complete_error(conn, &pgres, &error);
@@ -446,6 +453,7 @@ pq_abort_locked(connectionObject *conn, PGresult **pgres, char **error)
         return 0;
     }
 
+    conn->mark += 1;
     pq_clear_async(conn);
     retvalue = pq_execute_command_locked(conn, "ROLLBACK", pgres, error);
     if (retvalue == 0)
@@ -481,6 +489,8 @@ pq_abort(connectionObject *conn)
 
     pthread_mutex_unlock(&conn->lock);
     Py_END_ALLOW_THREADS;
+    
+    conn_notice_process(conn);
 
     if (retvalue < 0)
         pq_complete_error(conn, &pgres, &error);
@@ -537,6 +547,8 @@ pq_is_busy(connectionObject *conn)
     
     pthread_mutex_unlock(&(conn->lock));
     Py_END_ALLOW_THREADS;
+    
+    conn_notice_process(conn);
 
     return res;
 }
@@ -616,6 +628,8 @@ pq_execute(cursorObject *curs, const char *query, int async)
 
     pthread_mutex_unlock(&(curs->conn->lock));
     Py_END_ALLOW_THREADS;
+    
+    conn_notice_process(curs->conn);
 
     /* if the execute was sync, we call pq_fetch() immediately,
        to respect the old DBAPI-2.0 compatible behaviour */
@@ -807,23 +821,30 @@ _pq_copy_in_v3(cursorObject *curs)
        exception */
     PyObject *o;
     Py_ssize_t length = 0;
-    int error = 0;
+    int res, error = 0;
 
     while (1) {
         o = PyObject_CallMethod(curs->copyfile, "read",
-            CONV_CODE_PY_SSIZE_T, curs->copysize
-          );
+            CONV_CODE_PY_SSIZE_T, curs->copysize);
         if (!o || !PyString_Check(o) || (length = PyString_Size(o)) == -1) {
             error = 1;
         }
         if (length == 0 || length > INT_MAX || error == 1) break;
 
         Py_BEGIN_ALLOW_THREADS;
-        if (PQputCopyData(curs->conn->pgconn,
-                          PyString_AS_STRING(o),
-                          /* Py_ssize_t->int cast was validated above: */
-                          (int) length
-                         ) == -1) {
+        res = PQputCopyData(curs->conn->pgconn, PyString_AS_STRING(o),
+            /* Py_ssize_t->int cast was validated above */
+            (int) length);
+        Dprintf("_pq_copy_in_v3: sent %d bytes of data; res = %d",
+            (int) length, res);
+            
+        if (res == 0) {
+            /* FIXME: in theory this should not happen but adding a check
+               here would be a nice idea */
+        }
+        else if (res == -1) {
+            Dprintf("_pq_copy_in_v3: PQerrorMessage = %s",
+                PQerrorMessage(curs->conn->pgconn));
             error = 2;
         }
         Py_END_ALLOW_THREADS;
@@ -837,22 +858,37 @@ _pq_copy_in_v3(cursorObject *curs)
 
     Dprintf("_pq_copy_in_v3: error = %d", error);
 
-    if (error == 0 || error == 2)
-        /* 0 means that the copy went well, 2 that there was an error on the
-           backend: in both cases we'll get the error message from the
-           PQresult */
-        PQputCopyEnd(curs->conn->pgconn, NULL);
+    /* 0 means that the copy went well, 2 that there was an error on the
+       backend: in both cases we'll get the error message from the PQresult */
+    if (error == 0)
+        res = PQputCopyEnd(curs->conn->pgconn, NULL);
+    else if (error == 2)
+        res = PQputCopyEnd(curs->conn->pgconn, "error in PQputCopyData() call");
     else
-        PQputCopyEnd(curs->conn->pgconn, "error during .read() call");
+        res = PQputCopyEnd(curs->conn->pgconn, "error in .read() call");
 
-    /* and finally we grab the operation result from the backend */
     IFCLEARPGRES(curs->pgres);
-    while ((curs->pgres = PQgetResult(curs->conn->pgconn)) != NULL) {
-        if (PQresultStatus(curs->pgres) == PGRES_FATAL_ERROR)
-            pq_raise(curs->conn, curs, NULL);
-        IFCLEARPGRES(curs->pgres);
+    
+    Dprintf("_pq_copy_in_v3: copy ended; res = %d", res);
+    
+    /* if the result is -1 we should not even try to get a result from the
+       bacause that will lock the current thread forever */
+    if (res == -1) {
+        pq_raise(curs->conn, curs, NULL);
+        /* FIXME: pq_raise check the connection but for some reason even
+           if the error message says "server closed the connection unexpectedly"
+           the status returned by PQstatus is CONNECTION_OK! */
+        curs->conn->closed = 2;
     }
-
+    else {
+        /* and finally we grab the operation result from the backend */
+        while ((curs->pgres = PQgetResult(curs->conn->pgconn)) != NULL) {
+            if (PQresultStatus(curs->pgres) == PGRES_FATAL_ERROR)
+                pq_raise(curs->conn, curs, NULL);
+            IFCLEARPGRES(curs->pgres);
+        }
+    }
+   
     return error == 0 ? 1 : -1;
 }
 #endif
@@ -1122,6 +1158,8 @@ pq_fetch(cursorObject *curs)
 
     Dprintf("pq_fetch: fetching done; check for critical errors");
 
+    conn_notice_process(curs->conn);
+
     /* error checking, close the connection if necessary (some critical errors
        are not really critical, like a COPY FROM error: if that's the case we
        raise the exception but we avoid to close the connection) */
